// $Id: PageNumber.java 17152 2013-03-20 17:48:43Z mike $

package org.faceless.pdf2.viewer2.feature;

import org.faceless.pdf2.viewer2.*;
import org.faceless.pdf2.*;
import javax.swing.*;
import java.beans.*;
import java.util.*;
import java.awt.event.*;

/**
 * Create a widget that displays the current page number, and allows the user
 * to enter a new pagenumber for display.
 *
 * <div class="initparams">
 * The following <a href="../doc-files/initparams.html">initialization parameters</a> can be specified to configure this feature.
 * <table>
 * <tr><th>usePageLabels</th><td>If true (the default), display the {@link PDF#getPageLabel page label} for
 * the selected page if specified. If false, display the physical page number (the default behaviour
 * prior to 2.11.19).</td></tr>
 * <tr><th>alignment</th><td>Can be set to "right", "center" or "left" (the default) to position the page
 * number in the box</td></tr>
 * </table>
 *
 * <span class="featurename">The name of this feature is <span class="featureactualname">PageNumber</span></span>
 * <p><i>This code is copyright the Big Faceless Organization. You're welcome to use, modify and distribute it in any form in your own projects, provided those projects continue to make use of the Big Faceless PDF library.</i></p>
 * @since 2.8
 */
public class PageNumber extends ViewerWidget implements DocumentPanelListener, PropertyChangeListener {

    private JTextField field;
    private String[] pagelabels;
    private PDF activepdf;
    private boolean uselabels = true;

    public PageNumber() {
        super("PageNumber");

        field = new JTextField();
        field.setColumns(3);
        field.setFont(null);
        field.setEditable(true);
        field.setEnabled(false);
        field.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                String text = field.getText().trim();
                DocumentPanel docpanel = getViewer().getActiveDocumentPanel();
                if (pagelabels != null) {
                    int pn = -1;
                    for (int i=0;pn < 0 && i < pagelabels.length;i++) {
                        if (pagelabels[i].equals(text)) {
                            pn = i;
                        }
                    }
                    if (pn < 0) {               // Fallback to physical page number
                        try {
                            pn = Integer.parseInt(text) - 1;
                            if (pn < 0 || pn >= pagelabels.length) {
                                pn = -1;
                            }
                        } catch (NumberFormatException e2) { }
                    }
                    if (docpanel != null) {
                        if (pn >= 0) {
                            docpanel.setPageNumber(pn);
                        } else {
                            pn = docpanel.getPageNumber();
                            if (pn >= 0 && pn < pagelabels.length) {
                                field.setText(pagelabels[pn]);
                            }
                        }
                    }
                }
                if (field.hasFocus() && docpanel != null) {
                    docpanel.getViewport().requestFocusInWindow();
                }
            }
        });
        field.addFocusListener(new FocusAdapter() {
            public void focusGained(FocusEvent e) {
                field.selectAll();
            }
        });
        setComponent("Navigation.ltr", field);
    }

    public void initialize(PDFViewer viewer) {
        super.initialize(viewer);
        viewer.addDocumentPanelListener(this);
        String val = getFeatureProperty(viewer, "usePageLabels");
        if (val != null) {
            setUsePageLabels("true".equalsIgnoreCase(val));
        }
        String alignment = getFeatureProperty(viewer, "alignment");
        if ("right".equals(alignment)) {
            field.setHorizontalAlignment(JTextField.RIGHT);
        } else if ("center".equals(alignment)) {
            field.setHorizontalAlignment(JTextField.CENTER);
        } else {
            field.setHorizontalAlignment(JTextField.LEFT);
        }
    }

    /**
     * Set whether to display the "page labels" if defined on this PDF, or
     * whether to always display the physical page number.
     * @param uselabels if true, use the page labels if defined, otherwise use the physical page number
     * @see PDF#getPageLabel
     * @see ThumbnailPanel#setUsePageLabels
     * @since 2.11.19
     */
    public void setUsePageLabels(boolean uselabels) {
        this.uselabels = uselabels;
    }

    public void documentUpdated(DocumentPanelEvent event) {
        String type = event.getType();
        if (type=="pageChanged" || type=="activated") {
            if (type == "activated") {
                PDF pdf = event.getDocumentPanel().getPDF();
                if (pdf != null) {              // Shouldn't happen but...
                    activepdf = pdf;
                    activepdf.addPropertyChangeListener(this);
                    updatePageLabels(pdf);
                    field.setEnabled(true);
                }
            }
            int pagenumber = event.getDocumentPanel().getPageNumber();
            if (pagenumber >= 0 && pagelabels != null && pagenumber < pagelabels.length) {
                String label = pagelabels[pagenumber];
                if (!label.equals(field.getText())) {
                    field.setText(label);
                }
            }
        } else if (type == "deactivated") {
            if (activepdf != null) {
                activepdf.removePropertyChangeListener(this);
            }
            pagelabels = null;
            activepdf = null;
            field.setText("");
        }
    }

    public void propertyChange(PropertyChangeEvent e) {
        String name = e.getPropertyName();
        PDF pdf = activepdf;
        if (pdf != null && (name.equals("pages") || name.equals("pagelabels"))) {
            updatePageLabels(pdf);
        }
    }

    private void updatePageLabels(PDF pdf) {
        List l = new ArrayList();
        int numpages = pdf.getNumberOfPages();
        for (int i=0;i<numpages;i++) {
            String label = uselabels ? pdf.getPageLabel(i) : null;
            if (label == null) {
                label = Integer.toString(i + 1);
            }
            l.add(label);
        }
        pagelabels = (String[])l.toArray(new String[numpages]);
        DocumentPanel panel = getViewer().getActiveDocumentPanel();
        int pagenumber = panel == null ? -1 : panel.getPageNumber();
        if (pagenumber >= 0) {
            field.setText(pagelabels[pagenumber]);
        }
    }

}
